#include "flash.h"

#define FIS_entry_size (0x100/2)	//size of FIS Directory entry in words
#define FLASH_BLOCK_SIZE 0x10000

extern void diag_printf(const char *fmt, ...);
extern char *strcpy(char *s1, const char *s2);

#ifndef false
#define false 0
#endif

#ifndef true
#define true 1
#endif

struct fis_image_desc
{
  unsigned char name[16];	// Null terminated name
  unsigned long flash_base;	// Address within FLASH of image
  unsigned long mem_base;	// Address in memory where it executes
  unsigned long size;		// Length of image
  unsigned long entry_point;	// Execution entry point
  unsigned long data_length;	// Length of actual data
  unsigned char _pad[256 - (16 + 7 * sizeof(unsigned long))];
  unsigned long desc_cksum;	// Checksum over image descriptor
  unsigned long file_cksum;	// Checksum over image data
};


//=========================================================================================
// Local data used by these routines
const static int block_size = 0x10000;

static struct fis_image_desc img;

int memcmp(const void *s1, const void *s2, unsigned long n)
{
  const unsigned char *p1 = s1;
  const unsigned char *p2 = s2;
  
  while (n--)
    if (*p1++ != *p2++)
      return p1[-1] - p2[-1];
  
  return 0;
}

void *memset(void *s, int c, unsigned long n);
  
//---------------------------------------------------------------------------
void fis_init(void)
{
  int stat;

  // descriptor for the descriptor table itself
  memset(&img, 0, sizeof img);
  strcpy(img.name, "FIS_directory");
  img.flash_base = (unsigned long) 0x53f0000;
  img.mem_base = (unsigned long) 0x53f0000;
  img.size = block_size;

  if ((stat = (flash_erase_region((unsigned short *) 0x53f0000, FIS_entry_size))) != 0)
    {
      diag_printf("- Initialization failed: 0x%x\n", stat);
      return;
    }

  if ((stat = flash_program_region((unsigned short *) 0x53f0000, (unsigned short *)&img, FIS_entry_size)) != 0)
    {
      diag_printf("- Error writing image descriptors: 0x%x\n", stat);
      return;
    }
  
  if (memcmp((void*) 0x53f0000, &img, sizeof img))
    {
      diag_printf("verify failure\n");
      return;
    }
}

//----------    Simple XOR style checksum
unsigned long _cksum(unsigned long *buf, int len)
{
  unsigned long cksum = 0;
  // Round 'len' up to multiple of longwords
  len = (len + (sizeof(unsigned long) - 1)) / sizeof(unsigned long);
  while (len-- > 0)
    cksum ^= *buf++;
  return cksum;
}

//---------------------------------
void fis_create(void)
{
  int stat;
  unsigned long img_length, data_length, count;
  int statOK = true;
  unsigned long sourceChecksum;

  extern unsigned short __payloadStart[];
  extern unsigned short __payloadEnd[];
  extern volatile unsigned short __payloadDest[];
  extern unsigned short FlashDfltDir[];


  // count is actual data size (in 16-bit words)
  // data_length is actual data size in bytes
  // img_length is size of FLASH image, in bytes (rounded up length to FLASH block size)

  count = __payloadEnd - __payloadStart;
  data_length = 2 * count;
  img_length = ((data_length + block_size - 1) / block_size) * block_size;

  sourceChecksum = _cksum((unsigned long*)__payloadStart, data_length);
  
  if (data_length > img_length)
    {
      diag_printf("- Invalid FLASH image size/length combination\n");
      return;
    }

  //------------- Erase area to be programmed
  if ((stat = flash_erase_region(__payloadDest, count)) != 0)
    {
      diag_printf("- Can't erase region: 0x%x\n", stat);
      statOK = false;
    }

//  diag_printf("programming src=%x len=%d checksum=%08x\n", __payloadStart, data_length, _cksum((unsigned long*)__payloadStart, data_length));

  //------------- Now program it
  if ((stat = flash_program_region(__payloadDest, __payloadStart, count)) != 0)
    {
      diag_printf("- Can't program region: 0x%x\n", stat);
      statOK = false;
    }
  
  if (memcmp((void*)__payloadDest, __payloadStart, count*2))
    {
      diag_printf("- Verify failure\n");
    }

  //------------- Update directory information
  memset(&img, 0, sizeof img);
  strcpy(img.name, "default");
  img.flash_base = (unsigned long) 0x5030000;
  img.mem_base = (unsigned long) 0x0;
  img.entry_point = (unsigned long) 0x0;
  img.size = img_length;
  img.data_length = data_length;
  img.file_cksum = _cksum((unsigned long *) __payloadDest, img.data_length);

  if (img.file_cksum != sourceChecksum)
    {
      diag_printf("- Checksum verify error %08x %08x\n", sourceChecksum, img.file_cksum);
      statOK = false;
    }
  else
    {
      //------------ erase it
      if ((stat = flash_erase_region(FlashDfltDir, FIS_entry_size)) != 0)
        {
          diag_printf("- Error erasing: 0x%x\n", stat);
          statOK = false;
          // Don't try to program if the erase failed
        }
      else
        {
          //--------- program it
          if ((stat = flash_program_region(FlashDfltDir, (unsigned short *) &img, FIS_entry_size)) != 0)
            {
              diag_printf("- Error programming: 0x%x\n", stat);
              statOK = false;
            }
        }
    }

  if (statOK)
    diag_printf("+\n");
}

//===============================================================
