#include "flash.h"

//// According to Intel datasheet upper 8 bits of command should
// be all 0s or all 1s

#define FLASH_Read_ID      0x0090
#define FLASH_Read_Status  0x0070
#define FLASH_Clear_Status 0x0050
#define FLASH_Status_Ready 0x0080
#define FLASH_Program      0x0040
#define FLASH_Block_Erase  0x0020
#define FLASH_Suspend      0x00B0
#define FLASH_Resume       0x00D0
#define FLASH_Confirm      0x00D0
#define FLASH_Reset        0x00FF

#define FLASH_Config_Setup 0x0060
#define FLASH_Lock_Block   0x0001
#define FLASH_Unlock_Block 0x00D0
  
#define FLASH_BLOCK_SIZE        0x10000
#define FLASH_BOOT_BLOCK_SIZE    0x2000
#define FLASH_BOOT_BLOCK_COUNT   8

#define FLASH_Intel_code   0x89

#define ROM FlashBaseAddress

static unsigned long flashId;

extern void diag_printf(const char *fmt, ...);
extern void kickWatchdog(void);
extern volatile unsigned char EpldFlashWriteEnable;

static void flash_query(unsigned short *data)
{
  
  EpldFlashWriteEnable = 1;
  ROM[0] = FLASH_Read_ID;
  data[0] = ROM[0];  // Manufacturer code
  data[1] = ROM[1];  // Device identifier
  ROM[0] = FLASH_Reset;
  flashId = (data[0]<<16) | data[1];
  EpldFlashWriteEnable = 0;
}

int flash_hwr_init(void)
{
  unsigned short data[2];
  typedef int code_fun(unsigned short *);

  flash_query(data);
  if (data[0] == FLASH_Intel_code)
    diag_printf("\nFLASH:  Intel device type: 0x%04x\n", data[1]);
  else if (data[0] == 0x0020) 
    diag_printf("\nFLASH:  Numonyx device type 0x%04x\n",data[1]);
  else if (data[0] == 0x00B0)
    diag_printf("\nFLASH:  Sharp device type: 0x%04x\n", data[1]);
  else
    diag_printf("\nFLASH:  Unknown device found - vendor: 0x%x device: 0x%04x\n", data[0],  data[1]);
  return FLASH_ERR_OK;
}


// Map a hardware status to a package error
static int flash_hwr_map_error(int err)
{
   if (!(err & 0x007E007E))
     return FLASH_ERR_OK;

   diag_printf("FLASH:  ERROR\nFLASH:  Status Register Returned non-zero value:  0x%04x ", err);
   
   if (err & 0x00020002)
       diag_printf("-  Prog/Erase attempted on locked block ");
   if (err & 0x00040004) 
       diag_printf("-  Program suspend ");
   if (err & 0x00100010)
       diag_printf("-  Error programming a block ");
   if (err & 0x00200020)
       diag_printf("-  Error in block erase ");
   if (err & 0x00400040) 
       diag_printf("-  Erase suspend issued ");

   diag_printf("\n");

   if (err & 0x00400040) 
     return FLASH_ERR_ERASE_SUSPEND;
   else if (err & 0x00040004) 
     return FLASH_ERR_PROGRAM_SUSPEND;
   else if (err & 0x00020002)
     return FLASH_ERR_PROTECT;
   else if (err & 0x00100010)
     return FLASH_ERR_PROGRAM;
   else if (err & 0x00200020)
     return FLASH_ERR_ERASE;
   else
     return FLASH_ERR_HWR;
}

int get_status_timeout(unsigned t)
{
   unsigned long stat;
   while(((stat = (flush(),ROM[0])) & FLASH_Status_Ready) != FLASH_Status_Ready)
     {
	kickWatchdog();
	if (--t == 0) 
	  return FLASH_TIMEOUT;
     }
   return flash_hwr_map_error(stat);   
}

/*
 *  Lock/Unlock sector.  It appears the Intel flash part 0x88C5 requires
 *  sector lock/unlock, else the write will fail.  The part 0x8897 must 
 *  not be lock/unlock'd, else writes are very slooow.  What to do with 
 *  Sharp parts is unknow.
 */
static int lock_sector(volatile unsigned short *block)
{
   if (flashId != 0x008988C5 && flashId != 0x002088BB)
     return FLASH_ERR_OK;

   ROM[0] = FLASH_Clear_Status;
   ROM[0] = FLASH_Config_Setup;
   *block = FLASH_Lock_Block;
   return get_status_timeout(5000000);
}


static int unlock_sector(volatile unsigned short *block)
{
   if (flashId != 0x008988C5 && flashId != 0x002088BB)
     return FLASH_ERR_OK;

   ROM[0] = FLASH_Clear_Status;
   ROM[0] = FLASH_Config_Setup;
   *block = FLASH_Unlock_Block;
   return get_status_timeout(5000000);
}


int flash_erase_sector(volatile unsigned short *block)
{
   int s;

   diag_printf("FLASH:  Sector erase (addr: %08x)\n",block);
   EpldFlashWriteEnable = 1;
   unlock_sector(block);
   ROM[0] = FLASH_Clear_Status;
   *block = FLASH_Block_Erase;
   *block = FLASH_Confirm;
   s = get_status_timeout(5000000);
   ROM[0] = FLASH_Reset;
   lock_sector(block);
   EpldFlashWriteEnable = 0;
   return s;
}

int flash_erase_region(volatile unsigned short *addr, int count)
{
  volatile unsigned short *sectorBase = ROM;
  int sectorLen = FLASH_BOOT_BLOCK_SIZE;
  int status;
  
  diag_printf("FLASH:  Erase region (addr:%08x, num bytes: %d)\n", addr, count);
  
  while (sectorBase < addr + count)
    {
      if (sectorBase >= addr)
        {
          status = flash_erase_sector(sectorBase);
          if (status != FLASH_ERR_OK)
            return flash_hwr_map_error(status);
        }
      if (sectorBase >= ROM + FLASH_BOOT_BLOCK_COUNT * FLASH_BOOT_BLOCK_SIZE/(sizeof *ROM))
        sectorLen = FLASH_BLOCK_SIZE;
      
      sectorBase += sectorLen/(sizeof *ROM);
    }
  return FLASH_ERR_OK;
}

int flash_program_region(volatile unsigned short *addr, unsigned short *data, int count)
{
  int s = FLASH_ERR_OK;
  volatile unsigned short *last = (unsigned short*)0;
  int retryCount = 0;
  
  diag_printf("FLASH:  Programming... (target: %08x  source: %08x  bytes: %d)\n",addr, data, count * 2);
  
  EpldFlashWriteEnable = 1;
  ROM[0] = FLASH_Clear_Status;
  
  while (count > 0) 
    {
      // crossed an 8K boundary?
      if ((((unsigned long)addr)^((unsigned long)last)) & ~0x1fff)
        {
          if (last)
            lock_sector(last);
          unlock_sector(addr);
          last = addr;
        }

      *addr = FLASH_Program;
      *addr = *data;
      if ((s = get_status_timeout(10000000)))
        break;
      ROM[0] = FLASH_Reset;
      
      if (*addr != *data) 
        {
          diag_printf("%08x: %04x should be %04x\n",addr,*addr,*data);
          if (retryCount > 3)
            {
              diag_printf("FLASH:  ERROR \nFLASH:  Flash Verify Fail - Addr: 0x%x   Wrote: 0x%x  Read From Flash: 0x%x \n", addr, *data, *addr);
              s = FLASH_VERIFY_FAIL;
              break;
            }
          else
            {
              diag_printf("retrying\n");
              ++retryCount;
              continue;
            }
        }
      
      ++addr;
      ++data;
      --count;
      retryCount = 0;      
    }
  
  if (last)
    lock_sector(last);
  
  ROM[0] = FLASH_Reset;
  EpldFlashWriteEnable = 0;
  
  return s;
}
