/*********************************************************************
 *  iic.c
 *
 *  Methods to read and write iic memory
 *  Copied from $ECOS/hal/arm/dmrts/current/src/iic.c
 *  Edited to remove unnecessary functions other than getModelNumber()
 *
 *********************************************************************/

#include "iic.h"

#define Bit(n)          (1<<(n))

#define IIC_ACK         Bit(3)
#define IIC_Start       Bit(4)
#define IIC_Stop        Bit(5)
#define IIC_Restart     (Bit(5)|Bit(4))
#define IIC_BF          Bit(0)
#define IIC_Reset       Bit(7)

volatile unsigned int *iicCon = (volatile unsigned int *)0x7fff000;
volatile unsigned int *iicBuf = (volatile unsigned int *)0x7fff004;
volatile unsigned int *iicPrescale = (volatile unsigned int *)0x7fff008;

// Defined in linker script burn.ld
extern volatile unsigned PioMode;
extern volatile unsigned PioData;

static int iicInitialReset;

//  Does an initial reset of the IIC controller on the EEPROM part
void iicInit(void)
{
    *iicCon = IIC_Reset;
    *iicCon = IIC_BF;
    *iicPrescale = 100;
    *iicCon = 0;
    PioMode |= Bit(11);
}

//  Short delay.  Each nop is 22ns.
//  Inputs:  delay in microseconds
static void udelay(unsigned int usec)
{
    int i;

    for (i = 0; i < usec * 227; ++i)  {
        asm("nop");
        asm("nop");
    }
}

//  Waits up to 1 second for the BF flag to be set
//  Returns:  1 if OK, 0 if timeout occurred
static int waitForFlag(void)
{
    int count = 0;

    // allow up to a msec since the scheduler should not be running
    while ((*iicCon & IIC_BF) == 0) {
        if (count++ > 100) {      //  one msec
            iicInitialReset = 0;  // force a re-initialize of EEPROM
            return 0;
        }
        udelay(10);
    }
    return 1;
}

//  Reads IIC data, returns count bytes at p
//  Not thread safe, use "Rt" functions if scheduler is running
void iicRead(int addr, void *p, int count)
{
    unsigned char *buf = (unsigned char *)p;

    if (!iicInitialReset) {
        iicInit();
        iicInitialReset = 1;
        udelay(10);
    }

    *iicCon = IIC_Start+IIC_ACK;
    *iicBuf = 0xa0;
    if (!waitForFlag()) return;

    // need a delay here to avoid timing violation
    udelay(10);

    *iicBuf = addr;
    if (!waitForFlag()) return;

    // need a delay here to avoid timing violation
    udelay(10);

    *iicCon = IIC_Restart;
    *iicCon = IIC_Start+IIC_ACK;
    *iicBuf = 0xa1;
    if (!waitForFlag()) return;

    // need a delay here to avoid timing violation
    udelay(10);

    while (count) {
        // need a delay here to avoid timing violation
        udelay(1);

        if (count == 1)
            *iicCon = 0;
        else
            *iicCon = IIC_ACK;

        if (!waitForFlag()) return;

        // need a delay here to avoid timing violation
        udelay(1);

        PioData |= Bit(11);
        *buf++ = *iicBuf;
        PioData &= ~Bit(11);

        --count;
    }
    *iicCon = IIC_Stop;
}

void getModelNumber(unsigned long *m)
{
    iicRead(MODEL_NUMBER_ADDR, m, MODEL_NUMBER_SIZE);
}
